#ifndef __MORPION_HPP__
#define __MORPION_HPP__

#include "game.hpp"
#include <random>
#include <array>
#include <iostream>
#include <memory>

namespace game
{
	struct morpion_state
	{
		uint16_t cross_bitboard = 0;
		uint16_t circle_bitboard = 0;
		uint8_t total_moves = 0;
		bool first_player_win = false;
		bool second_player_win = false;
	};
	
	class morpion : public game<morpion_state>
	{
		public:
			morpion();
			morpion(const morpion& mor) = default;
			morpion& operator=(const morpion& mor) = default;
			bool end_of_game() const;
			int value(std::uint8_t player) const; //Returns if the player win, loose or nothing
			bool won(std::uint8_t player) const;
			bool lost(std::uint8_t player) const;
			bool draw(std::uint8_t player) const;
			uint8_t current_player() const;
			std::uint16_t number_of_moves() const;
			void play(std::uint16_t m);
			void undo(std::uint16_t m) {}
			std::string player_to_string(std::uint8_t player) const;
			std::string move_to_string(std::uint16_t m) const;
			std::string to_string() const;
			void playout(std::mt19937& engine, int max_depth = -1);
			std::set<int> to_input_vector() const;
			void from_input_vector(const std::set<int>& input);
			morpion_state get_state();
			void set_state(const morpion_state& state);
			std::shared_ptr<game<morpion_state>> do_copy() const;
			std::uint64_t hash(std::uint16_t m) const;
			std::uint64_t hash() const;
			
		private:
			inline void update_win();
			inline bool has_won(uint16_t bitboard);
			inline bool get(uint16_t bitboard, uint8_t i, uint8_t j) const;

			const uint8_t CROSS = 1;
			const uint8_t CIRCLE = 0;

			morpion_state state;

			static std::vector<std::vector<uint64_t>> cross_hash_values;
			static std::vector<std::vector<uint64_t>> circle_hash_values;
	};
	std::ostream& operator<<(std::ostream& os, const morpion& mor);
}

#endif
